import pandas as pd
import requests 
import zipfile  
import numpy as np
import urllib.request

path = 'https://raw.githubusercontent.com/victorncg/financas_quantitativas/main/Data%20Extraction/Stock%20Exchange/Index%20Composition/'
file = 'backend_index.py'


with urllib.request.urlopen(path + file) as response:
    py_content = response.read().decode('utf-8')

exec(py_content)



def _parse_ibov():

    try:

        url = "https://raw.githubusercontent.com/victorncg/financas_quantitativas/main/IBOV.csv"
        df = pd.read_csv(
            url, encoding="latin-1", sep="delimiter", header=None, engine="python"
        )
        df = pd.DataFrame(df[0].str.split(";").tolist())

        return df

    except:

        print("An error occurred while parsing data from IBOV.")



def _standardize_ibov():

    try:
        df = _parse_ibov()
        df.columns = list(df.iloc[1])
        df = df[2:][["Código", "Ação", "Tipo", "Qtde. Teórica", "Part. (%)"]]
        df.reset_index(drop=True, inplace=True)

        return df
    except:

        print("An error occurred while manipulating data from IBOV.")



def _standardize_sp500():
    """
    This function fetches the updated composition of the S&P 500 index. 
    
    Parameters
    ----------
    
    """

    table = pd.read_html("https://en.wikipedia.org/wiki/List_of_S%26P_500_companies")
    df = table[0]

    return df



def _adapt_index(
    index: object, assets: object = "all", mode: object = "df"
):
    """
    This function processes the data from the latest composition of either IBOV or S&P 500. 
    
    Parameters
    ----------
    index : choose the index to be returned, if IBOV or S&P 500
    ativos : you can pass a list with the desired tickets. Default = 'all'.
    mode: you can return either the whole dataframe from B3, or just the list containing the tickers which compose IBOV. Default = 'df'.
    
    """

    if index == "sp500":

        df = _standardize_sp500()

        if assets != "all":
            df = df[df["Symbol"].isin(assets)]

        if mode == "list":
            df = list(df.Symbol)

    else:

        df = return_index(index)

        if assets != "all":
            df = df[df["cod"].isin(assets)]

        if mode == "list":
            df = list(df.cod)
    
    return df



def index_composition(
    index: object, assets: object = "all", mode: object = "df"
):
    """
    This function captures the latest composition of either IBOV or S&P 500. It is updated every 4 months.
    
    Parameters
    ----------
    index : choose the index to be returned, if IBOV or S&P 500
    ativos : you can pass a list with the desired tickets. Default = 'all'.
    mode: you can return either the whole dataframe from B3, or just the list containing the tickers which compose IBOV. Default = 'df'.
    
    """

    df = _adapt_index(index, assets, mode)

    return df




def get_sectors(ticker:object = None):
    """
    This function get the B3 listed companies' classification into economic and activity sectors.
    You can leave the parameter 'ticker' empty if you wish to return all companies, or 
    specify a ticker without the numbers at the end, for example: "PETR"
    
    Parameters
    ----------
    ticker : you can pass a specific company's ticker. Default = None. 

    """    

    url_dataset = r"https://www.b3.com.br/data/files/57/E6/AA/A1/68C7781064456178AC094EA8/ClassifSetorial.zip"

    download = requests.get(url_dataset)
    with open('ClassifSetorial.zip', "wb") as dataset_B3:
        dataset_B3.write(download.content)
    arquivo_zip = zipfile.ZipFile('ClassifSetorial.zip')
    dataset = arquivo_zip.open(arquivo_zip.namelist()[0])

    df = pd.read_excel(dataset, header = 6)

    #Rename df columns:  
    df.rename(columns = {'LISTAGEM': 'CÓDIGO', 'Unnamed: 4':'SEGMENTO B3'}, inplace = True)
    #Create a new column:
    df['NOME NO PREGÃO'] = df['SEGMENTO'].copy()
    # Delete empty rows and duplicates:
    df.dropna(subset = ['NOME NO PREGÃO'], inplace = True)
    indexNames = df[df['SETOR ECONÔMICO'] == 'SETOR ECONÔMICO'].index
    df.drop(indexNames, inplace=True)
    df['SEGMENTO'] = np.where(df['CÓDIGO'].isna(),df['NOME NO PREGÃO'],pd.NA )    
    df['SETOR ECONÔMICO'] = df['SETOR ECONÔMICO'].ffill()
    df['SUBSETOR'] = df['SUBSETOR'].ffill()
    df['SEGMENTO'] = df['SEGMENTO'].ffill()
    df.dropna(subset = ['CÓDIGO'], inplace = True)
    
    #reset index:
    df.reset_index(drop=True, inplace=True)

    #
    df = df[['SETOR ECONÔMICO','SUBSETOR','SEGMENTO','NOME NO PREGÃO','CÓDIGO','SEGMENTO B3']]

    if ticker == None:
        df = df

    else:
        ticker = ticker[:4]
        df = df[df['CÓDIGO'] == ticker]

    return df


import time
import pandas as pd
import datetime

import functools
import logging

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


def _logging_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        logger.debug(f"function {func.__name__} called with args {signature}")
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            logger.exception(
                f"Exception raised in {func.__name__}. exception: {str(e)}"
            )
            raise e

    return wrapper


@_logging_error
def create_interval(start_date: object, end_date: object):

    """
    This function is responsible for receiving user input for start and end dates for data extraction and transforming it into YYYYMM format.
    
    :start_date: str
    :end_date: str
    """
    
    b = list()
    end = str(datetime.datetime.strptime(end_date, '%Y-%m-%d').year) + '{:02d}'.format(datetime.datetime.strptime(end_date, '%Y-%m-%d').month +1)

    for year in range(int(datetime.datetime.strptime(start_date, '%Y-%m-%d').year), int(datetime.datetime.strptime(end_date, '%Y-%m-%d').year)+1):

        for month in range(1,13):
            a = '{:02d}{:02d}'.format(year, month)
            if a == end:
                break
            b.append(a)

        if a == end:
            break

        year = year +1

    return b



@_logging_error
def extract_fund_data(dates_list: list, verbose: bool = False):
    start_time = time.time()

    # Initialize an empty list to store individual DataFrames
    dataframes = []
    # Iterate through the files in the folder
    for i in dates_list:

        url_pre = f'https://dados.cvm.gov.br/dados/FI/DOC/INF_DIARIO/DADOS/inf_diario_fi_{i}.zip'
        df_loc = pd.read_csv(url_pre, sep=';', compression='zip')
        s = datetime.datetime.strptime(i, "%Y%m")
        date = s.strftime('%B %Y')
        
        if verbose == True:
            print("Extraction of month",date,"finished")

        # Append the DataFrame to the list
        dataframes.append(df_loc)

    # Combine all DataFrames in the list into one large DataFrame
    combined_df = pd.concat(dataframes, ignore_index=True)

    if verbose == True:
        print("Process took %s seconds" % (time.time() - start_time))

    return combined_df


@_logging_error
def get_fund_data(start: object, end: object, cnpj:object = None, verbose: bool = False):

    """
    This function receives the raw data from function extract_fund_data and filters by date and CNPJ. If user doesn't specify CNPJ, it returns all funds.
  
    :start: str
    :end: str
    :cnpj: str
    :verbose: bool
    """
    interval = create_interval(start, end)
    data_fund = extract_fund_data(interval, verbose)
    data_fund['DT_COMPTC'] = pd.to_datetime(data_fund['DT_COMPTC'])
    mask = (data_fund['DT_COMPTC'] >= start) & (data_fund['DT_COMPTC'] <= end)
    data_fund = data_fund.loc[mask]

    if cnpj != None:
        data_fund = data_fund[data_fund['CNPJ_FUNDO'] == cnpj]

    else:
        data_fund = data_fund

    return data_fund.reset_index(drop=True)


@_logging_error
def get_funds_info(cnpj:object = None):

    """
    Fetches the registration data of investment funds from the Brazilian Securities and Exchange Commission (CVM).

    Parameters:
    - cnpj (str, optional): CNPJ (National Registry of Legal Entities) of the specific fund to retrieve information for. 
      If not provided, information for all funds is returned.

    Returns:
    - DataFrame: DataFrame containing the registration data of investment funds. Columns may include:
      - CNPJ: National Registry of Legal Entities.
      - DENOM_SOCIAL: Legal name of the company.
      - SIT: Status of the fund registration with the CVM.
      - GESTOR: Name of the fund manager.
      - ... (other relevant information)
    """
    
    url = "http://dados.cvm.gov.br/dados/FI/CAD/DADOS/cad_fi.csv"
    info = pd.read_csv(url, sep = ';', encoding = 'ISO-8859-1', low_memory=False)

    if cnpj != None:
        info = info[info['CNPJ_FUNDO'] == cnpj]

    return info.reset_index(drop = True)



import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.impute import SimpleImputer
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR
from xgboost import XGBRegressor
from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error


def get_categorical_data(df: pd.DataFrame):
    """
    This function is responsable for the identification of categorical and date features.

    :param df: Pandas DataFrame
    :type: pd.DataFrame
    :return: DataFrame with only categorical and date data types
    :rtype: DataFrame
    """

    return df.select_dtypes(include=['object'])


def get_numerical_data(df: pd.DataFrame):
    """
    This function is responsable for the identification of numerical features.

    :param df: Pandas DataFrame
    :type: pd.DataFrame
    :return: DataFrame with only numerical data types
    :rtype: DataFrame
    """

    return df.select_dtypes(include=['number', 'float64', 'int64'])


def handle_missing_data(df: pd.DataFrame, strategy_numerical='mean', strategy_categorical='most_frequent'):
    """
    This function is responsable for fill missings in dataFrame for numerical and categorical features with the substitution value passed by the user.

    Args:
        df: Pandas DataFrame
        strategy_numerical (str): Estratégia para lidar com valores ausentes em colunas numéricas.
                                   Opções são 'mean' (média), 'median' (mediana), 'most_frequent' (valor mais frequente).
        strategy_categorical (str): Estratégia para lidar com valores ausentes em colunas categóricas.
                                     Opções são 'most_frequent' (valor mais frequente) ou 'constant' (valor constante).

    Returns:
        tuple: Dados de treinamento e teste com valores ausentes tratados.
    """
    categorical_data = get_categorical_data(df)
    numerical_data = get_numerical_data(df)

    if len(categorical_data.columns) == 0:
        imputer_numerical = SimpleImputer(strategy=strategy_numerical)
        df_imputed_numerical = imputer_numerical.fit_transform(numerical_data)
        df_imputed = pd.DataFrame(df_imputed_numerical, columns=numerical_data.columns)

    else:
        # Cria um SimpleImputer para tratar valores ausentes em colunas numéricas
        imputer_numerical = SimpleImputer(strategy=strategy_numerical)
        df_imputed_numerical = imputer_numerical.fit_transform(numerical_data)

        # Cria um SimpleImputer para tratar valores ausentes em colunas categóricas
        imputer_categorical = SimpleImputer(strategy=strategy_categorical)
        df_imputed_categorical = imputer_categorical.fit_transform(categorical_data)

        # Concatena os dados imputados com as outras colunas
        df_imputed = pd.concat([pd.DataFrame(df_imputed_numerical, columns=numerical_data.columns),
                                pd.DataFrame(df_imputed_categorical, columns=categorical_data.columns)], axis=1)

    return df


def normalize_numerical_data(df: pd.DataFrame):
    """
    This function is responsable for normilize numerical data

    :param df: Pandas DataFrame
    :type: pd.DataFrame
    :return: DataFrame with numerical data normilized
    :rtype: DataFrame
    """

    numeric_columns = df.select_dtypes(include=['number', 'float64', 'int64']).columns
    scaler = StandardScaler(with_mean=True)
    df_normalized = scaler.fit_transform(df[numeric_columns])

    return pd.DataFrame(df_normalized, columns=numeric_columns)


def encode_nominal_categorical_data(df: pd.DataFrame, nominal_categorical_columns: list = None):
    """
    This function is responsable for encode nominal categorical data.

    :param df: Pandas DataFrame
    :type: pd.DataFrame
    :param nominal_categorical_columns: List of nominal categorical columns
    :type: List
    :return: DataFrame with nominal data encoded
    :rtype: DataFrame
    """

    # Encontra as colunas categóricas
    if nominal_categorical_columns is None:
        categorical_data = get_categorical_data(df)
        categorical_columns = categorical_data.columns
    else:
        categorical_columns = nominal_categorical_columns
        categorical_data = df[[categorical_columns]]

    # Codifica os dados categóricos usando one-hot encoding
    if len(categorical_data.columns) != 0:
        encoder = OneHotEncoder()
        df_encoded = encoder.fit_transform(df.iloc[:, categorical_columns])

        # Concatena os dados codificados com as outras colunas
        df_normalized = np.hstack([df_encoded, df.drop(df.columns[categorical_columns], axis=1)])

        return df_normalized
    else:
        return df


def encode_ordinal_categorical_data(df: pd.DataFrame, ordinal_categorical_columns: list = None):
    """
    This function is responsable for encode ordinal categorical data.

    :param df: Pandas DataFrame
    :type: pd.DataFrame
    :param ordinal_categorical_columns: List of ordinal categorical columns
    :type: List
    :return: DataFrame with ordinal data encoded
    :rtype: DataFrame
    """

    # Encontra as colunas categóricas
    if ordinal_categorical_columns is None:
        categorical_data = get_categorical_data(df)
        categorical_columns = categorical_data.columns
    else:
        categorical_columns = ordinal_categorical_columns
        categorical_data = df[[categorical_columns]]
    if len(categorical_data.columns) != 0:
        # Codifica os dados categóricos usando codificação ordinal
        encoder = OrdinalEncoder()
        df_encoded = encoder.fit_transform(df.iloc[:, categorical_columns])

        # Concatena os dados codificados com as outras colunas
        df_normalized = np.hstack([df_encoded, df.drop(df.columns[categorical_columns], axis=1)])

        return df_normalized
    else:
        return df


def regression_pipeline(df: pd.DataFrame, target_column, algorithm='random_forest',
                        ordinal_categorical_columns: list = None, nominal_categorical_columns: list = None,
                        strategy_numerical='mean', strategy_categorical='most_frequent', metrics: list = None):
    """
    Executes regression pipeline.

    :param df: Pandas DataFrame
    :type: pd.DataFrame
    :param target_column: data to predict
    :type: str
    :param algorithm: algorithm ('random_forest', 'gradient_boosting', 'mlp', 'svm', 'xgboost')
    :type: str
    :param ordinal_categorical_columns: List of ordinal categorical columns
    :type: List
    :param nominal_categorical_columns: List of nominal categorical columns
    :type: List
    :param strategy_numerical: Strategy to deal with nan values on numerical columns
    :type: str
    :param strategy_categorical: Strategy to deal with nan values on categorical columns
    :type: str
    :param metrics: list of metrics to return
    :type: List
    :return: y_pred with predicted target values
    :rtype: pd.Series
    :return: metrics_dict, dictionary of evaluation metrics
    :rtype: Dictionary

    Returns:
        model: Modelo treinado
        mse: Mean Squared Error do modelo
    """
    df = handle_missing_data(df, strategy_numerical, strategy_categorical)
    df = normalize_numerical_data(df)
    df = encode_nominal_categorical_data(df)
    df = encode_ordinal_categorical_data(df)

    # Dividir os dados em conjuntos de treinamento e teste
    y = df[target_column]
    X = df.drop(columns=[target_column])
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Treinar modelo de regressão
    models = {
        'random_forest': RandomForestRegressor(),
        'gradient_boosting': GradientBoostingRegressor(),
        'mlp': MLPRegressor(),
        'svm': SVR(),
        'xgboost': XGBRegressor()
    }
    if algorithm not in models:
        raise ValueError("Algoritmo de regressão não reconhecido.")

    model = models[algorithm]
    model.fit(X_train, y_train)

    y_pred = model.predict(X_test)
    metrics_dict = {}
    if metrics:
        if 'mse' in metrics:
            metrics_dict['mse'] = mean_squared_error(y_test, y_pred)
        if 'r2' in metrics:
            metrics_dict['r2'] = r2_score(y_test, y_pred)
        if 'mae' in metrics:
            metrics_dict['mae'] = mean_absolute_error(y_test, y_pred)
        if 'rmse' in metrics:
            metrics_dict['rmse'] = mean_squared_error(y_test, y_pred, squared=False)

    return y_pred, metrics_dict
